#include <crbprofiler.hpp>
#include <core_macros.h>
/*
	Author: Nelson Duarte
	Modified by Tyrghen
	
	Description:
	Loop stacked code/function with timing and conditional control
	Code and conditions are executed in non-scheduled environment
	
	Parameters:
	_this select 0: Action
	_this select 1: Parameters
	
	Parameters:
		id				Used to remove the element afterwards
		execution		The name of the function to execute
		timer			The number of units of time (X seconds or X frames)
		timerType		The type of units of time ("seconds","frames")
		conditionStart	The function name defining if the code can be executed
		conditionEnd	The function name defining if the code should be stopped
		executeOnce		A boolean defining if this is the last execution
		isSpawn			A boolean indicating if we should spawn the function or not.
						Note that if it still runs from a previous call, it won't be started again!
	
	Returns:
	NOTHING
	
	Examples:
	["initialize"] call mso_fnc_loop; 										//Initializes game loop (not required if AUTO_INITIALIZE is set to true)
	["isInitialized"] call mso_fnc_loop; 										//returns true if the loop is currently initialized
	["terminate"] call mso_fnc_loop; 										//Terminates game loop (not required if AUTO_TERMINATE is set to true)
	["itemAdd", ["uniqueId", "tmf_fnc_displayTime", 5]] call mso_fnc_loop;					//Hints time every five seconds
	["itemAdd", ["uniqueId", "tmf_fnc_displayTime", 120, "frames"]] call mso_fnc_loop; 				//Hints time every 120 frames
	["itemAdd", ["uniqueId", "tmf_fnc_displayTime", nil, nil, "tmf_fnc_startDisplay"]] call mso_fnc_loop; 	//Hints time every frame whenever tmf_fnc_startDisplay returns true
	["itemAdd", ["uniqueId", "tmf_fnc_displayTime", 5, "seconds", "tmf_fnc_startDisplay"]] call mso_fnc_loop; 	//Hints time every five seconds whenever tmf_fnc_startDisplay returns true
	["itemAdd", ["uniqueId", "tmf_fnc_displayTime"]] call mso_fnc_loop; 						//Hints time every frame
	["itemRemove", ["uniqueId"]] call mso_fnc_loop; 								//Remove item from loop with id "uniqueId"
*/

//Defines
#define AUTO_INITIALIZE false
#define AUTO_TERMINATE false
#define EXECUTION_INTERVAL_FRAMES 10

CRBPROFILERSTART("mso_fnc_loop")

//Parameters
private ["_action", "_parameters"];
_action		= [_this, 0, "", [""]] call BIS_fnc_param;
_parameters	= [_this, 1, []] call BIS_fnc_param;

switch (toLower _action) do {
	case (toLower "initialize") : {
		//If initialized already, we terminate current instance and initialize a new one
		//All data will be lost
		if (!isNil { missionNameSpace getVariable QGVAR(fnc_loop_initialized) }) then {
			//Terminate
			["terminate"] call FUNC(loop);
		};
		
		//Queue
		missionNameSpace setVariable [QGVAR(fnc_loop_queue), []];
		
		//The current amount of elapsed frames
		missionNameSpace setVariable [QGVAR(fnc_loop_frames), 0];
		
		//The current time
		missionNameSpace setVariable [QGVAR(fnc_loop_time), time];
		
		//On each frame event handler
		missionNameSpace setVariable [QGVAR(fnc_loop_onEachFrame), [QUOTE(FUNC(loop)), "oneachframe", {
			//Function is called every frame
			["oneachframe"] call FUNC(loop);
		}] call BIS_fnc_addStackedEventHandler];
		
		//Flag
		missionNameSpace setVariable [QGVAR(fnc_loop_initialized), true];
		
		//Log
		"Initialized" call BIS_fnc_log;
	};
	
	case (toLower "terminate") : {
		//Terminate on each frame loop
		[missionNameSpace getVariable QGVAR(fnc_loop_onEachFrame), "oneachframe"] call BIS_fnc_removeStackedEventHandler;
		
		//Clear variables
		missionNameSpace setVariable [QGVAR(fnc_loop_initialized), nil];
		missionNameSpace setVariable [QGVAR(fnc_loop_queue), nil];
		missionNameSpace setVariable [QGVAR(fnc_loop_frames), nil];
		missionNameSpace setVariable [QGVAR(fnc_loop_time), nil];
		missionNameSpace setVariable [QGVAR(fnc_loop_onEachFrame), nil];
		
		//Log
		"Terminated" call BIS_fnc_log;
	};
	
	case (toLower "onEachFrame") : {
		//Initialized?
		if (isNil { missionNameSpace getVariable QGVAR(fnc_loop_initialized) }) exitWith {
			//Error
			["Action (%1) requires initialization", _action] call BIS_fnc_error;
		};
		private ["_lastframes","_frames"];
		_lastframes = missionNameSpace getVariable [QGVAR(fnc_loop_lastframes),(-EXECUTION_INTERVAL_FRAMES)];
		//Update frames and time
		_frames = (missionNameSpace getVariable QGVAR(fnc_loop_frames)) + 1;
		missionNameSpace setVariable [QGVAR(fnc_loop_frames), _frames];
		missionNameSpace setVariable [QGVAR(fnc_loop_time), time];
		
		if ((_frames - _lastframes) > EXECUTION_INTERVAL_FRAMES) then {
			//Go through all items in queue
			{
				private "_parameters";
				_parameters = _x;
				_parameters set [8, _forEachIndex];
				
				//Execute item
				["itemExecute", _parameters] call FUNC(loop);
			} forEach (missionNameSpace getVariable QGVAR(fnc_loop_queue));
			missionNameSpace setVariable [QGVAR(fnc_loop_lastframes),_frames];
		};
		//If queue has no more items we terminate loop
		if (AUTO_TERMINATE && { count (missionNameSpace getVariable QGVAR(fnc_loop_queue)) < 1 }) then {
			//Log
			"Queue does not have anymore items" call BIS_fnc_log;
			
			//Terminate
			["terminate"] call FUNC(loop);
		};
	};
	
	case (toLower "itemExecute") : {
		//Initialized?
		if (isNil { missionNameSpace getVariable QGVAR(fnc_loop_initialized) }) exitWith {
			//Error
			["Action (%1) requires initialization", _action] call BIS_fnc_error;
		};
		
		//Parameters
		private ["_id", "_code", "_params", "_timer", "_timerType", "_conditionStart", "_conditionEnd", "_executeOnce", "_lastExecutionTime", "_index","_isSpawn","_start","_end","_handle"];
		_id 				= GETVAL(_parameters, 0, "");
		_code 				= GETVAL(_parameters, 1, "");
		_params				= GETVAL(_parameters, 2, []);
		_timer				= GETVAL(_parameters, 3, 0);
		_timerType 			= GETVAL(_parameters, 4, "seconds");
		_conditionStart		= GETVAL(_parameters, 5, "");
		_conditionEnd		= GETVAL(_parameters, 6, "");
		_executeOnce 		= GETVAL(_parameters, 7, false);
		_isSpawn 			= GETVAL(_parameters, 8, false);
		_lastExecutionTime 	= GETVAL(_parameters, 9, time);
		_index				= GETVAL(_parameters, 10, 0);
		_handle				= GETVAL(_parameters, 11, false);
		
		//The condition, depends on timerType
		private "_conditionTimer";
		_conditionTimer = if (_timerType == "frames") then {
			(missionNameSpace getVariable QGVAR(fnc_loop_frames)) >= _lastExecutionTime + _timer;
		} else {
			time >= _lastExecutionTime + _timer;
		};

		//Should we execute the item?
		if (_conditionTimer) then {
			_start = true;
			if (_conditionStart != "") then {
				_start = _parameters call (missionNameSpace getVariable [_conditionStart,{false}]);
			};
			if (_start) then {
				_end = false;
				//Process item
				// We ignore items which are not strings designating a function
				if (IS_STRING(_code)) then {
					if (_isSpawn) then {
						// We only execute the code if the previous iteration already ended
						if (typename _handle == typename true || {scriptDone _handle}) then {
							_handle = _parameters spawn (missionNameSpace getVariable [_code,{}]);
						};
					} else {
						_parameters call (missionNameSpace getVariable [_code,{}]);
					};
				} else {
					// This automatically removes non string items
					_end = true;
				};
				
				// We check for the ending condition if there is a function provided
				if (!_end && _conditionEnd != "") then {
					_end = _parameters call (missionNameSpace getVariable [_conditionEnd,{true}]);
				};
				
				//Should only execute once or end condition met?
				if (_executeOnce || _end) then {
					//Remove item
					["itemRemove", _id] call FUNC(loop);
				} else {
					//Queue
					private "_queue";
					_queue = missionNameSpace getVariable QGVAR(fnc_loop_queue);
					
					//Item
					private "_item";
					_item = _queue select _index;
					
					//Update timer/frames
					if (_timerType == "frames") then {
						_item set [8, missionNameSpace getVariable QGVAR(fnc_loop_frames)];
					} else {
						_item set [8, time];
					};
					if (_isSpawn) then {
						_item set [11, _handle];
					};
					//Modify item in queue
					_queue set [_index, _item];
					
					//Store
					missionNameSpace setVariable [QGVAR(fnc_loop_queue), _queue];
				};
			};
		};
	};
	
	case (toLower "itemAdd") : {
		//Public
		private ["_id", "_code", "_params", "_timer", "_timerType", "_conditionStart", "_conditionEnd", "_executeOnce","_isSpawn","_runNow","_lastExecution","_array"];
		_id 			= GETVAL(_parameters, 0, "");
		_code 			= GETVAL(_parameters, 1, "");
		_params			= GETVAL(_parameters, 2, []);
		_timer			= GETVAL(_parameters, 3, 0);
		_timerType 		= GETVAL(_parameters, 4, "seconds");
		_conditionStart	= GETVAL(_parameters, 5, "");
		_conditionEnd	= GETVAL(_parameters, 6, "");
		_executeOnce 	= GETVAL(_parameters, 7, false);
		_runNow 		= GETVAL(_parameters, 8, false);
		_isSpawn 		= GETVAL(_parameters, 9, false);
		
		//Initialize loop if not initialized already
		if (isNil { missionNameSpace getVariable QGVAR(fnc_loop_initialized) }) then {
			//Error
			if (!AUTO_INITIALIZE) exitWith {
				["Action (%1) requires initialization", _action] call BIS_fnc_error;
			};
			
			//Auto initialize
			["initialize"] call FUNC(loop);
		};
		if (!(IS_STRING(_id)) || {_id == ""}) exitWith { ["Invalid type or content for element id: %1", typename _id] call BIS_fnc_error; };
		if (!(IS_STRING(_code))) exitWith { ["Invalid type for execution function: %1", typename _code] call BIS_fnc_error; };
		if (!(IS_STRING(_conditionStart))) exitWith { ["Invalid type for start condition function: %1", typename _conditionStart] call BIS_fnc_error; };
		if (!(IS_STRING(_conditionEnd))) exitWith { ["Invalid type for end condition function: %1", typename _conditionEnd] call BIS_fnc_error; };
		if (isNil { missionNameSpace getVariable _code }) exitWith {
			["Unknown execution function on this instance: %1", _code] call BIS_fnc_error;
		};
		if ((_conditionStart != "") && {isNil { missionNameSpace getVariable _conditionStart }}) exitWith {
			["Unknown start function on this instance: %1", _conditionStart] call BIS_fnc_error;
		};
		if ((_conditionEnd != "") && {isNil { missionNameSpace getVariable _conditionEnd }}) exitWith {
			["Unknown end function on this instance: %1", _conditionEnd] call BIS_fnc_error;
		};
		if (!(IS_STRING(_timerType))) then { _timerType = "seconds"; };
		_timerType = switch(toLower(_timerType)) do {
			case "seconds": { "seconds" };
			case "secondes": { "seconds" };
			case "second": { "seconds" };
			case "seconde": { "seconds" };
			case "sec": { "seconds" };
			case "frames": { "frames" };
			case "frame": { "frames" };
			case "frm": { "frames" };
			default { "seconds" };
		};
		if (!(IS_BOOL(_executeOnce))) then { _executeOnce = false; };
		if (!(IS_BOOL(_runNow))) then { _runNow = false; };
		if (!(IS_BOOL(_isSpawn))) then { _isSpawn = false; };
		
		_lastExecution = time;
		// We change the last Execution time to make sure it runs on the next frame
		if (_runNow) then { _lastExecution = time - (_timer max 1); };
		
		_array = [_id, _code, _array, _timer, _timerType, _conditionStart, _conditionEnd, _executeOnce, _isSpawn, _lastExecution, 0];
		if (_isSpawn) then {
			// In the special case of Spawn scripts, we add a terminated handle
			_array set [count _array,(0 spawn {})];
		};
		
		//Queue
		private "_queue";
		_queue = missionNameSpace getVariable QGVAR(fnc_loop_queue);
		//Add item to queue
		_queue set [count _queue, _array];
		//Store queue
		missionNameSpace setVariable [QGVAR(fnc_loop_queue), _queue];
		
		//Log
		["Item added: Id (%1), Code (%2), Timer (%3), Timer Type (%4), Condition Start (%5), Condition End (%6), Execute Once (%7), Run Now (%8/%9), Is Spawn (%10)", _id, _code, _timer, _timerType, _conditionStart, _conditionEnd, _executeOnce,_runNow,_lastExecution,_isSpawn] call BIS_fnc_logFormat;
	};
	
	case (toLower "itemRemove") : {
		//Initialized?
		if (isNil { missionNameSpace getVariable QGVAR(fnc_loop_initialized) }) exitWith {
			//Error
			["Action (%1) requires initialization", _action] call BIS_fnc_error;
		};
		
		//Parameters
		private ["_id"];
		_id = [_parameters, 0, "", [""]] call BIS_fnc_param;
		
		//Queue
		private "_queue";
		_queue = missionNameSpace getVariable QGVAR(fnc_loop_queue);
		
		//Go through queue and find matching id
		{
			//The current item id
			private "_itemId";
			_itemId = [_x, 0, "", [""]] call BIS_fnc_param;
			
			//Match found?
			if (_id == _itemId) then {
				_queue set [_forEachIndex, "delete"];
			};
		} forEach _queue;
		
		//Now we remove all dirty items from the queue
		while { "delete" in _queue } do {
			_queue = _queue - ["delete"];
		};
		
		//Store queue
		missionNameSpace setVariable [QGVAR(fnc_loop_queue), _queue];
		
		//Log
		["Item removed: Id (%1)", _id] call BIS_fnc_logFormat;
	};
	
	case (toLower "isinitialized") : {
		//Initialized?
		(isNil { missionNameSpace getVariable QGVAR(fnc_loop_initialized) })
	};
	
	case DEFAULT {
		//Error
		["Invalid action: %1", _action] call BIS_fnc_error;
	};
};

CRBPROFILERSTOP